"""
Protoc compiler wrapper CLI (using `fire`_)

.. _fire:
    https://python-fire.readthedocs.io/en/latest/
"""
from __future__ import annotations

import distutils.log
import distutils.spawn
import functools
import os
import pathlib
import re
import subprocess
import warnings
from typing import List, Optional, Dict

import itertools
import sys

from . import find_proto_files
from .options import CompileOption

package_regex = r'(\w+)((?:\.\w+)*)'
check_packages = functools.partial(re.match, package_regex)
_DEFAULT = object()


class Compiler:
    """
    Wrapper around protobuf compiler.

    See ``compile-proto OPTIONS`` for all possible compilation options.
    See ``compile-proto`` (no args) and ``compile-proto --help`` for more help
    """

    OPTIONS = [o.formatted_argument for o in CompileOption if o.formatted_argument]
    """
    Supported compile options, multiple options may be specified.
    """

    def __init__(self, protoc=distutils.spawn.find_executable('protoc')):
        self.protoc = protoc or self._find_protoc()

    def _find_protoc(self) -> Optional[str]:
        """
        Searches for a protoc executable respecting the PROTOC
        environment variable

        Returns:
            path to protoc executable or None
        """
        protoc = os.environ.get('PROTOC', distutils.spawn.find_executable('protoc'))
        if not protoc:
            warnings.warn(f"protoc is not found in $PATH."
                          " Please install it or set the PROTOC environment variable")
        return protoc

    def call(self, *proto_files,
             includes: str = None,
             protohelp=False,
             dry_run=False,
             **options):
        """
        Just a wrapper around the `protoc` compiler.

        See ``compile-proto call --help`` to print help of protoc command
        See ``compile-proto call -- --help`` for additonal help with the call command
        See ``compile-proto protoc`` for the path of the used protoc compiler executable

        Args:
            includes: Directories to use as includes
            protohelp: Use this flag to pass ``--help`` to protoc invocation
            options: Will be passed as flags to the protoc command (only ``--flag`` syntax supported, no single dash)
            dry_run: If True, don't do anything except printing the command
        """
        if 'help' in options:
            print("To pass `--help` to protoc, use flag `--protohelp`. "
                  "Make sure you use only options with --{option} syntax when calling the protoc compiler "
                  "from `compile-proto call`. i.e. options with single dashes are not supported."
                  " If you want to get more help about the `call` command, use `compile-proto -- --help`")
            return

        if protohelp:
            options['help'] = protohelp

        protoc_command: List[str] = [os.fspath(self.protoc)]  # protoc executable
        protoc_command += [f"-I{os.fspath(include)}" for include in includes or ()]  # includes
        protoc_command += [f'--{k}={v}' for k, v in options.items()]  # protoc arguments
        protoc_command += [os.fspath(f) for f in proto_files]  # .proto files
        result = 0

        distutils.log.debug(" ".join(str(c) for c in protoc_command))
        if not dry_run:
            print(' '.join(protoc_command))
            result = subprocess.call(protoc_command)

        if result != 0:
            sys.exit(result)

    def compile(self, *protoc_files,
                options=_DEFAULT,
                quiet=False,
                plugin_params: str = '',
                output=os.getcwd(),
                **kwargs) -> Optional[str]:
        """
        Compile for given options, see ``compile-proto compile -- --help``

        Args:
            quiet: Don't print output from protoc plugin if possible
            output: output directory (default: working directory)
            options: one or multiple options see `compile-proto OPTIONS` default: [py]
            protoc_files: files to compile, passed through to protoc
            plugin_params: mapping of additional parameters for the protoc plugin (not the compiler itself)
            kwargs: Passed to protoc invocation, see compile-proto call -- --help.

        Returns:
            the output path
        """
        params = (plugin_params,) if plugin_params else ()

        if options is _DEFAULT:
            options = ['py']

        if not options:
            warnings.warn("No options specified, no compilation will take place.")
            return

        for option in CompileOption.from_string_list(options).disjunct:
            params += ('quiet',) if quiet else ()

            protoc_args = {
                f'{option.protoc_plugin_name}_out': option.format_out(output, *params)
            }

            distutils.log.info(f"Compiling with {protoc_args}")
            self.call(*protoc_files, **kwargs, **protoc_args)


class Rewriter:
    """
    Rewrite proto files to force python package structure
    which mirrors protobuf package declarations.
    """

    _IMPORT = re.compile(r'^import "((?:\w+/)*)(\w+\.proto)";$', flags=re.MULTILINE)
    _PACKAGE = re.compile(r'^package {pkg};$'.format(pkg=package_regex), flags=re.MULTILINE)

    def __init__(self,
                 root_package: str,
                 output_root: os.PathLike = './'):

        self._contents: Optional[Dict[pathlib.Path, str]] = None
        self._roots: Optional[Dict[pathlib.Path, pathlib.Path]] = None
        self.root_package = root_package
        self.output_root = output_root

    def read(self, *dirs: os.PathLike):
        """
        Reads .proto files from directories, manipulate them with the other commands afterwards.
        """
        dirs = [pathlib.Path(s) for s in dirs]
        found = {s: find_proto_files(s) for s in dirs if s.is_dir()}
        no_proto_dirs = [s for s in dirs if not s in found or not found[s]]
        assert not no_proto_dirs, f"Check path[s] {no_proto_dirs}: Not a directory or does not contain a .proto file."

        # invert dictionary to lookup parents
        self._roots = {path: root for root, paths in found.items() for path in paths}
        self._contents = {path: path.read_text(encoding='utf-8') for path in self._roots}
        return self

    @property
    def root_package(self):
        """
        root package for generated directory structure

        Returns:
            str: root package which is enforced
        """
        return self._root_package

    @root_package.setter
    def root_package(self, value):
        if value is None:
            return

        valid = check_packages(value)
        assert valid, f"{value} not a valid package name, please use only .-separated package names"

        self._root_package = value

    @property
    def output_root(self):
        """
        Root directory of new directory structure.
        """
        return self._out_root

    @output_root.setter
    def output_root(self, value):
        self._out_root = pathlib.Path(value)

    def _get_package(self, path):
        relative_path = path.relative_to(self._roots[path])
        package = self._PACKAGE.search(self._contents[path])
        return self._fix_package(package) if package else '.'.join(relative_path.parts)

    def _fix_package(self, match):
        # replace first part of package with self._package
        # applying fix_package multiple times is ok since it doesn't change the package
        root_pkg, sub_pkgs = match.groups()
        sub_pkgs = sub_pkgs[1:].split('.')  # skip first char, since it's a dot anyways
        pkg_iterator = itertools.dropwhile(lambda p: p in self._root_package, itertools.chain([root_pkg], sub_pkgs))
        return '.'.join(itertools.chain([self._root_package], pkg_iterator))

    @property
    def calculated_packages(self):
        """
        Returns:
            dict: Dictionary mapping file contents to package names
        """
        return {p: self._get_package(p) for p in self._contents or ()}

    def _fix_import(self, root: pathlib.Path, match):
        path, file = match.groups()
        import_package = self.calculated_packages.get(root / path / file)
        return '/'.join(itertools.chain(import_package.split('.'), [file])) if import_package else None

    def _fix_import_declaration(self, root, match):
        return f'import "{self._fix_import(root, match)}";'

    def fix_imports(self):
        """
        Modify internal file content representation to match internal file locations, so that imports are importing
        the right files.

        Returns:
            Rewriter: Reference to self to chain commands with `fire`_

        """
        process_imports = {p: {statement: self._fix_import(self._roots[p], statement)}
                           for p, content in self._contents.items()
                           for statement in self._IMPORT.finditer(content)}

        failed = {p: imports for p, imports in process_imports.items() if any(v is None for v in imports.values())}

        if any(failed):
            warnings.warn('\n'.join("Can't resolve imports {imports} from file {path}".format(
                path=path, imports=[k.group(0) for k, v in imports.items() if v is None]
            ) for path, imports in failed.items()) +
                          " Make sure the respective files are also included for rewriting")
            return self

        self._contents = {f: self._IMPORT.sub(functools.partial(self._fix_import_declaration, self._roots[f]), content)
                          for f, content in self._contents.items()}
        return self

    def fix_packages(self):
        """
        Modify internal content representation such that ``package`` declarations in ``.proto`` source files match the
        internal directory structure

        Returns:
            Rewriter: Reference to self to chain commands with `fire`_
        """
        def make_regex(declared_package):
            root_package = declared_package[1]
            sub_packages = declared_package[2].replace('.', r'\.') if declared_package[2] else ""
            regex_str = r"(?:{}\.)?({})({})".format(self._root_package, root_package, sub_packages)
            return re.compile(regex_str)

        unique_package_regexes = set(map(
            make_regex,
            itertools.chain.from_iterable(
                self._PACKAGE.finditer(content) for content in self._contents.values()
            )
        ))

        for regex in unique_package_regexes:
            self._contents = {
                f: regex.sub(self._fix_package, content) for f, content in self._contents.items()
            }

        return self

    def content(self, filename: os.PathLike) -> str:
        """
        View the internal file representations

        Args:
            filename: Input file path

        Returns:
            str: contents of internal file representation

        """
        return self._contents.get(pathlib.Path(filename), f"File for filename {filename} not found. Available"
                                                          f" filenames: {', '.join(map(str, self._contents))}")

    def write(self, dry_run=True):
        """
        Write internal content representation to :attr:`.output_root` according to
        internal :attr:`package mapping <.calculated_packages>`

        Args:
            dry_run (bool): if True don't actually write outputs.

        Returns:
            Rewriter: Reference to self to chain commands with `fire`_
        """

        for file, content in self._contents.items():
            out_dir = self.output_root / '/'.join(self._get_package(file).split('.'))
            out_dir.mkdir(parents=True, exist_ok=True)

            with open(out_dir / file.name, 'w') as output:
                if dry_run:
                    print(f"Fake writing {file.name} in {out_dir} since dry_run is {dry_run}.")
                    continue

                output.write(content)


def check_fire():
    try:
        import fire
    except ImportError:
        distutils.log.error("Can't use CLI for compiler if python-fire is not installed!"
                            "Did you install the package with [cli]?")
    else:
        return fire


def compile_proto():
    """
    Entry point for CLI.

        `fire`_ :class:`Compiler`

    """

    fire = check_fire()
    if fire:
        fire.Fire(Compiler)


def rewrite_proto():
    """
    Entry point for CLI

        `fire`_ :class:`Rewriter`
    """
    fire = check_fire()
    if fire:
        fire.Fire(Rewriter)



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import datetime
import os
import pathlib
import sys

sys.path.insert(0, os.path.abspath('.'))
sys.path.insert(0, os.path.abspath('../../src'))

try:
    import importlib.metadata as importlib_metadata
except ImportError:
    import importlib_metadata

metadata = importlib_metadata.metadata('codestare-msg-compiler')

# -- Project information -----------------------------------------------------

project = metadata['Name']
copyright = f'{datetime.datetime.now().year}, {metadata["Author"]}'
author = 'Maximilian Schmidt'
url = metadata['Home-Page']
# The full version, including alpha/beta/rc tags
release = metadata['version']

# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'sphinx.ext.autodoc',
    'sphinxcontrib.napoleon',
    'sphinx.ext.intersphinx',
    'sphinxcontrib.programoutput',
    'sphinx.ext.viewcode',
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = []

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = 'alabaster'

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']

html_css_files = [
    'css/custom.css'
]

html_sidebars = {
    '**': [
        'about.html',
        'navigation.html',
        'relations.html',
        'searchbox.html',
    ]
}

html_theme_options = {
    'description': metadata['Summary'],
    'github_user': 'saggitar',
    'github_repo': 'codestare-msg-compiler',
    'page_width': '75%'
}

autodoc_preserve_defaults = True
# autodoc_typehints = 'description'
# autodoc_typehints_description_target = 'documented'
# autodoc_class_signature = 'separated'
# autodoc_class_signature = 'mixed'

autodoc_default_options = {
    'member-order': 'bysource',
    'members': True,
    'undoc-members': False,
    'show-inheritance': True,
}

autodoc_inherit_docstrings = False
napoleon_include_special_with_doc = True

intersphinx_mapping = {
    'python': ('https://docs.python.org/3', None),
    'plus': ('https://proto-plus-python.readthedocs.io/en/latest', None),
    'proto': ('https://googleapis.dev/python/protobuf/latest', None),
}

import shutil
example_project_build_dirs = [
    pathlib.Path("example_project/src/py/namespace/proto/v1"),
    pathlib.Path("./example_project/build")
]
for directory in example_project_build_dirs:
    directory.mkdir(parents=True, exist_ok=True)
    shutil.rmtree(directory)
    directory.mkdir()



from .v1 import (
    Service,
    ServiceList,
    ServiceReply,
    ServiceRequest,
    TopicSubscription,
    RegexSubscription,
    Error,
    ErrorList,
    Success,
    SuccessList,
)
__all__ = (
    "Service",
    "ServiceList",
    "ServiceReply",
    "ServiceRequest",
    "TopicSubscription",
    "RegexSubscription",
    "Error",
    "ErrorList",
    "Success",
    "SuccessList",
)



#!/usr/bin/env python

"""
Googles own python plugin implementation for protoc does not compile files with relative imports, like
e.g. `betterproto`_ (related `issue`_). Several third party tools try to "fix" this by either pre- or post-processing
the generated files, or using a custom ``protoc`` plugin (like `betterproto`_).

Conversely, `betterproto's <betterproto>`_ relative imports only succeed when all sources are compiled simultaneously
(all proto files as arguments to `protoc` invocation, otherwise files might be overwritten).
This would break support for build tools like cmake / make, allowing parallel compilation of ``.proto`` files
i.e. it seems not possible to implement in the official plugin.

Google itself develops an `alternative protobuf module <protoplus>`_ which aims to use protocol buffers with more
idiomatic python code but has no official compiler plugin as of now. A third party plugin `plusplugin`_ loosely based on
the `mypy plugin <mypy-protobuf>`_ is available (experimental).

.. _issue:
   https://github.com/protocolbuffers/protobuf/issues/1491

.. _betterproto:
   https://github.com/danielgtaylor/python-betterproto

.. _protoplus:
   https://github.com/googleapis/proto-plus-python

.. _plusplugin:
   https://github.com/saggitar/proto-plus-plugin

.. _mypy-protobuf:
   https://github.com/nipunn1313/mypy-protobuf
"""
import contextlib
import distutils.cmd
import distutils.errors
import distutils.log
import enum
import filecmp
import fnmatch
import functools
import importlib.resources
import itertools
import os
import os.path as op
import pathlib
import re
import sys
import tempfile
from abc import ABC
from typing import List, Optional

import setuptools.command.build_py
import setuptools.command.egg_info

from . import find_proto_files, has_module
from .compile import Compiler, Rewriter
from .options import CompileOption

PathList = Optional[List[pathlib.Path]]


@contextlib.contextmanager
def compare_files():
    """
    Context manager to monkey patch :func:`distutils.file_util.copy_file` to only copy files which are not equal
    according to :func:`filecmp.cmp`.

    Yields:
        None

    Example:
        Use the context manager with methods implicitly using :func:`distutils.file_util.copy_file`::

            class CustomCommand(distutils.cmd.Command):
                def copy_not_equal(new, old):
                    with compare_files():
                        self.copy_tree(new, old)
    """
    import distutils.file_util as fu
    orig = fu.copy_file

    @functools.wraps(orig)
    def wrapper(src, dst, *args, verbose=1, **kwargs):
        if pathlib.Path(dst).exists() and filecmp.cmp(src, dst):
            if verbose >= 1:
                distutils.log.debug("not copying %s (output up-to-date)", src)
            return dst, 0

        return orig(src, dst, *args, verbose=verbose, **kwargs)

    fu.copy_file = wrapper
    yield
    fu.copy_file = orig


class PathCommand(distutils.cmd.Command, ABC):
    """
    Abstract command class supporting verification of options representing path and directory lists
    """
    def ensure_path_list(self, option):
        """
        Ensure option is list of paths

        Args:
            option (str): name of option

        """
        val = getattr(self, option)
        if val is None:
            return

        if not isinstance(val, list) or not all(isinstance(o, pathlib.Path) for o in val):
            self.ensure_string_list(option)
            val = [pathlib.Path(s) for s in getattr(self, option)]

        not_exist = [p for p in val if not p.exists()]
        if any(not_exist):
            raise distutils.errors.DistutilsOptionError(
                f"Paths {', '.join(str(o.absolute()) for o in not_exist)} don't exist.")

        setattr(self, option, val)

    def ensure_dir_list(self, option):
        """
        Ensure option is list of directories

        Args:
            option (str): name of option

        """
        self.ensure_path_list(option)
        val = getattr(self, option)
        if val is None:
            return

        not_dir = [p for p in val if not p.is_dir()]
        if any(not_dir):
            raise distutils.errors.DistutilsOptionError(
                f"Paths {', '.join(str(o.absolute()) for o in not_dir)} are not directories.")


class CompileBase(PathCommand):
    """
    Base class for setuptools commands handling compilation of ``.proto`` files.

    Attributes:

        includes (List[pathlib.Path]): Directories containing ``.proto`` files to pass as includes to ``protoc``
        output (pathlib.Path): compilation output directory
        files (List[pathlib.Path], optional): actual list of ``.proto`` files to compile
        proto_package (pathlib.Path, optional): if :obj:`.files` is missing, search this path for ``.proto`` files
        force (bool): Force compilation / copying of all generated files disregarding changed contents. Defaults to False.
        dry_run (bool): Don't copy generated files to build dir, only print list of generated files. Defaults to False.
        options (List[str]): List of :attr:`option flags <codestare.compiletools.compile.Compiler.OPTIONS>`
            to trigger compilation flavors
        protoc (pathlib.Path): Path to ``protoc`` executable. If not supplied, ``protoc`` needs to be in ``$PATH``
        plugin_params (str, optional): Some ``protoc`` plugins support additional parameters
    """
    description = "Compile proto files"

    user_options = [
        ('protoc=', None, 'protoc compiler location'),
        ('output=', None, 'Output directory for compiled files'),
        ('includes=', None, 'Include directories for .proto files'),
        ('force', 'f', "forcibly build everything (ignore file timestamps)"),
        ('files=', None, 'Protobuf source files to compile'),
        ('plugin_params=', None, 'parameters passed to protoc plugin'),
        ('options=', 'o', f"Options for compilation, possible values are "
                          f"{CompileOption.ALL.disjunct}"
                          f" (default)")
    ]
    """
    User options for :class:`distutils.cmd.Command`. See :meth:`CompileBase.run` for more info about how they are used
    in particular.
    
    Each option is used to supply the corresponding attribute: for more information about e.g. ``user_options['force']``
    refer to :attr:`.force`.
    
    User options are passed between build steps. If not supplied specifically for this build command, the following
    mapping occurs (options specified multiple times will take the first existing default value):
    
    ============= ===========================================
    option        default ``{command_name}.{option_name}``
    ============= ===========================================
    includes      rewrite_proto.outputs
    output        compile_proto.build_lib
    includes      compile_proto.include_proto
    proto_package compile_proto.proto_package
    force         compile_proto.force
    dry_run       compile_proto.force
    ============= ===========================================
    
    """

    @contextlib.contextmanager
    def redirect_build_dir(self):
        """
        Context manager to redirect build directory to temporary directory.
        This is the easiest way to only "build" changed protobuf modules:
        We build to temp dir, then copy to real build dir if files are different.

        Note:

            There is no way to tell the ``protoc`` compiler to only build files that have changed / updated sources.
            This is typically the task of ``make`` or some other build script, and all files passed to ``protoc`` will
            be built (in theory the plugin could also handle this, but none of the default plugins do). This means
            they will have new file generation timestamps and will

            - probably be marked as changed in the version control system
            - be copied to the build directory by the python build toolchain

            Because of this, we need to compare the file contents instead of the timestamps, even when building to
            a temporary directory, see :func:`compare_files`.


            This context manager changes :attr:`.output` as a side effect

        Yields:
            Tuple[pathlib.Path, pathlib.Path]: old build dir, new build dir

        """
        original = self.output
        tf = None
        if not self.force:
            executable = pathlib.Path(sys.executable)
            suffix = '_'.join(p.replace('.', '_') for p in executable.parts if not p == executable.anchor)
            tf = tempfile.TemporaryDirectory(suffix=suffix,
                                             prefix=__name__.replace('.', '_'))
            self.output = tf.name

        yield original, self.output
        if tf is not None:
            tf.cleanup()

        self.output = original

    def run(self):
        """
        Calls :meth:`Compiler.compile` with :attr:`.user_options` as keyword arguments.

        This triggers a compilation of :attr:`.files` using the :attr:`.protoc` executable, passing :attr:`.includes`
        as includes (``-I`` flag) and :attr:`.plugin_params`, using the plugin and default options for :attr:`.options`
        (see :mod:codestare.compiletools.options) with output directory :attr:`.output`.

        If :attr:`.force` is not set, only generated files that don't exist with the same contents in :attr:`.output`
        will be copied from the temporary build directory.

        If :attr:`.dry_run` is set, no files will be copied whatsoever, and only the ``protoc`` invocation will be
        shown.

        """
        compiler = Compiler(protoc=self.protoc)

        with self.redirect_build_dir() as (old, temp):
            args = {k: v for k, v in vars(self).items() if k in ['options', 'output', 'includes', 'plugin_params']}
            args['quiet'] = not self.distribution.verbose
            compiler.compile(*self.files, **args)

            with compare_files():
                self.copy_tree(temp, old)

        for command in self.get_sub_commands():
            self.run_command(command)

    def finalize_options(self) -> None:
        self.set_undefined_options('rewrite_proto',
                                   ('outputs', 'includes'))

        self.set_undefined_options('compile_proto',
                                   ('build_lib', 'output'),
                                   ('include_proto', 'includes'),
                                   ('proto_package', 'proto_package'),
                                   ('force', 'force'),
                                   ('dry_run', 'dry_run')
                                   )

        self.ensure_dir_list('includes')
        self.ensure_dirname('output')
        self.ensure_string_list('options')
        self.ensure_filename('protoc')
        self.ensure_string('plugin_params')

        if self.files is not None:
            self.ensure_path_list('files')
        elif self.proto_package:
            self.files = find_proto_files(*[include / '/'.join(self.proto_package.split('.'))
                                            for include in self.includes])
        else:
            self.files = find_proto_files(*self.includes)

    def initialize_options(self) -> None:
        self.protoc = None
        self.output = None
        self.includes: PathList = None
        self.files: PathList = None
        self.dry_run = None
        self.options = None
        self.force = None
        self.proto_package = None
        self.plugin_params = None


class CompileProtoPython(CompileBase):
    """
    Command to compile with compilation option :attr:`~codestare.compiletools.options.CompileOption.PYTHON_BASIC`
    and generate ``__init__.py`` files with :class:`GenerateInits`
    """

    description = "compile python protobuf modules (google plugin)"
    user_options = CompileBase.user_options[:-1]

    def initialize_options(self) -> None:
        super().initialize_options()
        self.options = 'py'

    sub_commands = [
        ('generate_inits', None)
    ]


class CompileProtoMypy(CompileBase):
    """
    Command to compile with compilation option :attr:`~codestare.compiletools.options.CompileOption.PYTHON_MYPY`,
    to generate ``.pyi`` stubs.
    """
    description = "compile stub files for python protobuf modules (mypy plugin)"
    user_options = CompileBase.user_options[:-1]

    def initialize_options(self) -> None:
        super().initialize_options()
        self.options = 'mypy'


class CompileBetterproto(CompileBase):
    """
    Command to compile with compilation option :attr:`~codestare.compiletools.options.CompileOption.PYTHON_BETTER_PROTO`
    and generate ``__init__.py`` files with :class:`GenerateInits`
    """
    description = "compile alternative python protobuf modules (betterproto plugin)"
    user_options = CompileBase.user_options[:-1]

    def initialize_options(self) -> None:
        super().initialize_options()
        self.options = 'better'

    sub_commands = [
        ('generate_inits', None)
    ]

class CompileProtoPlus(CompileBase):
    """
    Command to compile with compilation option :attr:`~codestare.compiletools.options.CompileOption.PYTHON_BETTER_PROTO`
    and generate ``__init__.py`` files with :class:`GenerateInits`

    Note:
        Although generating ``__init__.py`` files via the :class:`GenerateInits` setuptools command is supported for
        :class:`CompileProtoPlus`, the ``protoc`` plugin used in this compilation handles generation of init files.

        It is recommended to not use the :class:`generate_inits command<GenerateInits>` or at least not to use ``force``
        to avoid overwriting the files.

    """
    description = "compile alternative python protobuf modules (protoplus plugin)"
    user_options = CompileBase.user_options[:-1]

    def initialize_options(self) -> None:
        super().initialize_options()
        self.options = 'plus'

    sub_commands = [
        ('generate_inits', None)
    ]


class CompileProto(PathCommand):
    """
    Triggers compilation of all available python flavors that are turned on via :attr:`.flavor`.
    If :attr:`.proto_package` and :attr:`.include_proto` are supplied, enforce package structure according to
    :attr:`.proto_package` by rewriting included source files with
    :class:`rewrite subcommand <codestare.compiletools.cmd.RewriteProto>`

    Attributes:
        include_proto (List[pathlib.Path]): root dir for proto files
        proto_package (str): parent package that will be enforced for protobuf modules
        build_lib (os.path.PathLike): output directory for protobuf library
        flavor (str, optional): flavor of compiled code
        force (bool, optional): forcibly build everything (ignore file timestamps)
        dry_run (bool, optional): don't do anything but show protoc commands
    """

    description = "compile protobuf files with [all] available python plugins"

    class Flavor(enum.Enum):
        """
        Possible Flavors for python compilations
        """
        MYPY = 'mypy'
        BASIC = 'python'
        BETTER = 'better'
        PLUS = 'plus'

    flavors = {s.value: s for s in Flavor}
    """
    Mapping of options to enums
    """

    user_options = [
        ('include-proto', None, 'root dir for proto files'),
        ('proto-package', None, 'parent package that will be enforced for protobuf modules'),
        ('build-lib', None, 'output directory for protobuf library'),
        ('flavor', None, f'flavor of compiled code, one of {",".join(flavors)}'),
        ('force', 'f', "forcibly build everything (ignore file timestamps)"),
        ('dry-run', None, 'don\'t do anything but show protoc commands')
    ]
    """
    User options for :class:`distutils.cmd.Command` behaviour.

    Each option is used to supply the corresponding attribute: for more information about e.g. ``user_options['force']``
    refer to :attr:`.force`.

    User options are passed between build steps. If not supplied specifically for this build command, the following
    mapping occurs (options specified multiple times will take the first existing default value):

    ============= ===========================================
    option        default ``{command_name}.{option_name}``
    ============= ===========================================
    build_lib     build_py_proto.build_lib
    include_proto build_py_proto.include_proto
    force         build_py_proto.force
    dry_run       build_py_proto.dry_run
    ============= ===========================================

    """

    def initialize_options(self) -> None:
        self.include_proto = None
        self.proto_package = None
        self.build_lib = None
        self.flavor = None
        self.force = None
        self.dry_run = None

    def finalize_options(self) -> None:
        self.set_undefined_options('build_py_proto',
                                   ('build_lib', 'build_lib'),
                                   ('dry_run', 'dry_run'),
                                   ('force', 'force'),
                                   ('include_proto', 'include_proto'))

        self.ensure_path_list('include_proto')
        self.ensure_string('proto_package')
        self.ensure_string('flavor')
        if self.flavor is not None:
            if self.flavor not in self.flavors:
                raise distutils.errors.DistutilsOptionError(
                    f"Only possible options for flavor are {','.join(self.flavors)}")
            self.flavor = self.flavors[self.flavor]

    def run(self):
        """
        Trigger all compilations and add compiled packages to distribution if missing.

        Note:
            If python packages are generated which are not already part of the distribution (through setting
            ``packages`` option in ``setup.py`` or ``setup.cfg`` as usual) this command will issue a warning

        """
        if self.include_proto is not None:
            self.announce(f"Including *.proto files from path[s] {self.include_proto}", distutils.log.INFO)
        else:
            self.announce(f"No *.proto files specified, not compiling.", distutils.log.INFO)

        for command in self.get_sub_commands():
            self.run_command(command)

        proto_package_path = self.proto_package.split('.') if self.proto_package else ()
        compiled_packages = ['.'.join([self.proto_package, p])
                             for p in setuptools.find_packages(op.join(self.build_lib, *proto_package_path))]

        if compiled_packages:
            self.announce(f"built protobuf packages: {compiled_packages}", distutils.log.INFO)

        missing = [p for p in compiled_packages or () if p not in self.distribution.packages]
        if missing:
            distutils.log.warn(f"protobuf packages {', '.join(missing)} are missing "
                               f"from setup.cfg / setup.py, but have been compiled.")
            self.distribution.packages += missing

    def mypy_rule(self):
        return (has_module('mypy', 'mypy_protobuf')
                and self.flavor == self.Flavor.MYPY
                and self.include_proto)

    def better_proto_rule(self):
        return (has_module('betterproto')
                and self.flavor == self.Flavor.BETTER
                and self.include_proto)

    def proto_plus_rule(self):
        return (has_module('codestare.proto')
                and self.flavor == self.Flavor.PLUS
                and self.include_proto)

    def basic_python_rule(self):
        return (self.flavor == self.Flavor.BASIC
                and self.include_proto)

    def rewrite_rule(self):
        return (self.proto_package is not None
                and self.include_proto)

    sub_commands = [
        ('rewrite_proto', rewrite_rule),
        ('compile_python', basic_python_rule),
        ('compile_betterproto', better_proto_rule),
        ('compile_mypy', mypy_rule),
        ('compile_protoplus', proto_plus_rule),
    ]


class RewriteProto(PathCommand):
    """
    Setuptools command to rewrite ``.proto`` sources in a way that produces a specified python package structure.

    Attributes:
        proto-package (str): parent package that will be enforced for protobuf modules
        inplace (bool): write output to compile_proto include directory
        use_build (bool): write output to build_lib directory [default]

    """
    description = "rewrite protobuf inputs to fake specific package structure (experimental)"

    user_options = [
        ('proto-package=', None, 'parent package that will be enforced for protobuf modules'),
        ('inplace', None, 'write output to compile_proto include directory'),
        ('use-build', None, 'write output to build_lib directory [default]'),
    ]

    boolean_options = ['inplace']
    negative_opt = {'use-build': 'inplace'}

    def initialize_options(self) -> None:
        self.proto_package = None
        self.inplace = 0
        self.dry_run = None
        self.inputs = None
        self.outputs = None

    def finalize_options(self) -> None:
        self.set_undefined_options('compile_proto',
                                   ('dry_run', 'dry_run'),
                                   ('build_lib', 'outputs'),
                                   ('proto_package', 'proto_package'),
                                   ('include_proto', 'inputs'),
                                   )

        if self.inplace:
            self.outputs = self.inputs

        self.ensure_string('proto_package')
        self.ensure_dir_list('inputs')
        self.ensure_dir_list('outputs')

        if self.inputs is None:
            raise distutils.errors.DistutilsOptionError(f"Can't rewrite proto files without inputs")

        if self.outputs is None:
            raise distutils.errors.DistutilsOptionError(f"Can't rewrite proto files without outputs")

        if len(self.inputs) != len(self.outputs):
            raise distutils.errors.DistutilsOptionError(f"can't rewrite proto files from {self.inputs}: "
                                                        f"wrong number of outputs ({len(self.outputs)})")

        self.announce(f"Enforcing python package {self.proto_package} for compiled modules.", distutils.log.INFO)

    def run(self) -> None:
        rewriter = Rewriter(root_package=self.proto_package)

        for input, output in zip(self.inputs, self.outputs):
            rewriter.read(input)
            rewriter.output_root = output
            rewriter.fix_imports()
            rewriter.fix_packages()
            rewriter.write(dry_run=self.dry_run)


class GenerateInits(PathCommand):
    """
    Setuptools command to generate missing ``__init__.py`` files in generated python package tree.
    """
    class _Styles(enum.Enum):
        FANCY = 'fancy'
        WILDCARD = 'wildcard'
        EMPTY = 'empty'

    styles = {s.value: s for s in _Styles}
    description = "generate (better) __init__.py files for protobuf modules"

    user_options = [
        ('packages', None, 'generate for these packages only'),
        (
            'recursive', None,
            'if set, you only need to specify parent packages, all subpackages will also be considered'),
        ('no-recursive', None, 'only the exact packages specified in `packages` are considered. [default]'),
        ('import_style', None, f'One of: {", ".join(styles)}. See documentation for details about generated inits'),
    ]

    boolean_options = ['recursive']
    negative_opt = {'no-recursive': 'recursive'}

    def initialize_options(self) -> None:
        self.recursive = 0
        self.import_style = None
        self.package_root = None
        self.packages = None
        self.force = None

    def finalize_options(self) -> None:
        self.set_undefined_options('compile_python',
                                   ('force', 'force'),
                                   ('output', 'package_root'))

        self.ensure_string_list('packages')
        self.ensure_dirname('package_root')
        self.ensure_string('import_style')

        if self.package_root is None:
            raise distutils.errors.DistutilsOptionError(f"Can't generate inits without "
                                                        f"package root (no output from `compile_python`?)")

        if self.import_style is not None:
            if self.import_style not in self.styles:
                raise distutils.errors.DistutilsOptionError(
                    f"Only supported values for import_style are: {', '.join(self.styles)}")
            else:
                self.import_style = self.styles[self.import_style]

    def run(self) -> None:
        """
        Generate recursive init files with wildcard imports for a package.
        """

        def is_package(path: pathlib.Path):
            return path.is_dir() and list(path.glob('**/__init__.py'))

        if self.packages is None:
            root = pathlib.Path(self.package_root)
            available = (f"'{p.name}'" for p in filter(is_package, filter(pathlib.Path.is_dir, root.glob('*'))))
            self.announce(f"no packages specified -> skipping. "
                          f"(possible packages found in {self.package_root}: "
                          f"{', '.join(available) or 'No packages found'})", distutils.log.INFO)

            return

        search_dirs = (
            pathlib.Path(self.package_root) / op.join(*package.split('.'))
            for package in self.packages or ()
        )

        searches = (
            p.glob(f"{'**' if self.recursive else '*'}/")
            for p in search_dirs
        )

        skipped = []

        for package in itertools.chain(*searches):
            init: pathlib.Path = package / '__init__.py'
            name = '.'.join(package.relative_to(self.package_root).parts)
            if init.exists() and not self.force:
                skipped += [name]
                distutils.log.debug(f"Not generating {init}, already existing. Use --force")
                continue

            with init.open('w', encoding='utf-8') as f:
                if self.import_style == self._Styles.FANCY:
                    f.write(importlib.resources.read_text(__package__, 'init_template'))
                if self.import_style == self._Styles.WILDCARD:
                    modules = [p for p in init.parent.glob('*.py') if not p.name.startswith('_')]
                    f.write('\n'.join(f"from .{m.stem} import *" for m in modules))
                if self.import_style == self._Styles.EMPTY:
                    pass

        created = [p for p in self.packages or () if p not in skipped]
        if created:
            self.announce(f"Generated __init__.py files for "
                          f"python packages {created}"
                          f"{' recursively' if self.recursive else ''}", distutils.log.INFO)
        elif self.packages:
            self.announce(f"__init__.py files in {self.packages} already present, not generated", distutils.log.INFO)
        else:
            self.announce(f"No init files generated for {self.package_root}", distutils.log.INFO)


class UbiiBuildPy(setuptools.command.build_py.build_py):
    def __getattr__(self, item):
        if item == 'user_options':
            return setuptools.command.build_py.build_py.user_options + CompileProto.user_options

        return setuptools.command.build_py.build_py.__getattr__(self, item)

    def initialize_options(self) -> None:
        super().initialize_options()
        self.include_proto = None

    def finalize_options(self) -> None:
        super().finalize_options()

        if self.include_proto == "included":
            # if .proto files are included in data files
            proto_dirs = [self.get_package_dir(package) for package, _, _, filenames in self.data_files
                          if any(re.match(fnmatch.translate('*.proto'), f) for f in filenames)]
            parent_paths = op.commonprefix(proto_dirs)
            if parent_paths:
                self.include_proto = [os.fspath(p) for p in parent_paths]

    def run(self):
        for command in self.get_sub_commands():
            self.run_command(command)

        super().run()

    sub_commands = [
        ('compile_proto', None),
    ]


def write_package(cmd: setuptools.command.egg_info.egg_info, basename, filename, force=False):
    compile_command = cmd.get_finalized_command('compile_proto')
    proto_plus_cmd = (cmd.get_finalized_command('compile_protoplus')
                      if 'compile_protoplus' in compile_command.get_sub_commands()
                      else None)

    value = getattr(compile_command, 'proto_package', None)
    params = getattr(proto_plus_cmd, 'plugin_params', '')
    if 'package' in params:
        value = params[params.index('package'):].split(',')[0].replace('package', value)

    if value:
        cmd.write_or_delete_file('proto package name', filename, value, force)



import importlib
import pathlib
import typing

from itertools import chain


def has_module(*modules: str):
    """
    Try importing module, catch ImportError

    Args:
        *modules: module names

    Returns: True if module is importable else False
    """

    try:
        for name in modules:
            importlib.import_module(name)
    except ImportError:
        return False
    return True


def find_proto_files(*paths: pathlib.Path, recursive: bool = True) -> typing.List:
    """
    Return relative paths of .proto files in paths

    Args:
        *paths: We :func:`~pathlib.Path.glob` the files in the paths
        recursive: check recursively in all paths if True

    Returns:
       List of unique paths of found .proto files, relative to search path
    """
    searcher = chain.from_iterable(
        p.glob(f"{'**/' if recursive else ''}*.proto") for p in paths
    )
    found = list(dict.fromkeys(searcher))
    return found



"""
Options for compilation are represented using :class:`.CompileOption`
"""
import operator
from enum import Flag, auto
from functools import reduce
from itertools import chain
from typing import List, Union

CompileOptionStr = Union['Compiler.CompileOption', str]


class CompileOption(Flag):
    """
    Describes available compile options
    """
    # empty docstrings for enum values to include them in documentation
    #:
    JAVA = auto()
    #:
    JSLIBRARY = auto()
    #:
    JSINDIVIDUAL = auto()
    #:
    CSHARP = auto()
    #:
    CPP = auto()
    #:
    PYTHON_BETTER_PROTO = auto()
    #:
    PYTHON_MYPY = auto()
    #:
    PYTHON_PROTOPLUS = auto()
    #:
    PYTHON_BASIC = auto()
    #:
    JAVASCRIPT = JSINDIVIDUAL | JSLIBRARY
    #:
    PYTHON = PYTHON_MYPY | PYTHON_BETTER_PROTO | PYTHON_PROTOPLUS | PYTHON_BASIC
    #:
    ALL = JAVA | JAVASCRIPT | CSHARP | CPP | PYTHON

    @property
    def protoc_plugin_name(self):
        if self == self.JAVA: return 'java'
        if self == self.JSLIBRARY: return 'js'
        if self == self.JSINDIVIDUAL: return 'js'
        if self == self.CSHARP: return 'csharp'
        if self == self.CPP: return 'cpp'
        if self == self.PYTHON_BETTER_PROTO: return 'python_betterproto'
        if self == self.PYTHON_MYPY: return 'mypy'
        if self == self.PYTHON_BASIC: return 'python'
        if self == self.PYTHON_PROTOPLUS: return 'proto-plus'
        # combined options don't have plugins

    @classmethod
    def from_str(cls, value) -> 'CompileOption':
        """
        Convert a string to an enum value

        Args:
            value (str): valid values are defined by :attr:`.arguments`

        Returns:
            CompileOption: Enum flag
        """
        matching = reduce(operator.and_, [o for o in cls if value in o.arguments])

        if not matching:
            raise ValueError(f"{value} is no valid option.")

        return matching

    @classmethod
    def from_string_list(cls, values):
        """
        Convert a string list to an enum value (combined flag for all values in list)

        Args:
            values (List[str]): list of values according to :attr:`.arguments`

        Returns:
            CompileOption: combined flag

        """
        return reduce(operator.and_, [cls.from_str(v) for v in values])

    @property
    def arguments(self) -> List:
        """
        Arguments that can be used to trigger this compilation
        """
        arguments = []
        if self & self.JAVA: arguments += ['java']
        if self & self.CSHARP: arguments += ['cs']
        if self & self.CPP: arguments += ['cpp']
        if self & self.PYTHON_MYPY: arguments += ['mypy']
        if self & self.PYTHON_BETTER_PROTO: arguments += ['better']
        if self & self.PYTHON_PROTOPLUS: arguments += ['plus']
        if self & self.PYTHON_BASIC: arguments += ['py']
        # other options have no associated arguments

        # combined options with special arguments:
        if self == self.ALL:
            return ['all']
        if self == self.JAVASCRIPT:
            return ['js']

        return arguments

    def format_out(self, output, *params, **kw_params):
        return ':'.join(filter(None, (self.parameters(*params, **kw_params), output)))

    @property
    def formatted_argument(self):
        return f"[{','.join(self.arguments)}] {self}" if len(self.arguments) == 1 else None

    @property
    def disjunct(self) -> List['CompileOption']:
        """
        Return all options which are not combinations of other options
        """
        return [o for o in type(self) if o & self and not o.is_composite]

    @property
    def is_composite(self):
        """
        Like normal integer flags, individual flags are all powers of 2,
        see https://docs.python.org/3/library/enum.html#flag
        combined flags are not, so we test that with ``o.value & (o.value - 1)``
        """
        return self.value & (self.value - 1)

    def parameters(self, *args, **kwargs):
        """
        some options can use request parameters for the plugin, this is e.g.
        how the javascript plugin implements "library / binary" compilation.

        - :attr:`.JSLIBRARY` adds ``library='protobuf_library',binary``
        - :attr:`.JSINDIVIDUAL` adds ``import_style='commonjs',binary``
        - :attr:`.PYTHON_PROTOPLUS` adds ``readable_imports``
        """
        args = set(args)

        if self.is_composite:
            raise ValueError("parameters only make sense for individual options.")

        if self == self.JSLIBRARY:
            kwargs['library'] = 'protobuf_library'
            args.add('binary')

        if self == self.JSINDIVIDUAL:
            kwargs['import_style'] = 'commonjs'
            args.add('binary')

        return ','.join(chain((f'{k}={v}' for k, v in kwargs.items()), args))

    def __str__(self):
        if len(self.disjunct) > 1:
            return " | ".join(str(o) for o in self.disjunct)

        if self == self.PYTHON_BETTER_PROTO: return 'python with `better_proto` plugin'
        if self == self.PYTHON_MYPY: return 'mypy python stubs'
        if self == self.PYTHON_MYPY: return 'python with `proto-plus` plugin'
        if self == self.PYTHON_BASIC: return 'python'
        if self == self.JSLIBRARY: return 'javascript as library'
        if self == self.JSINDIVIDUAL: return 'javascript individual'
        # other options can just use the name
        return self.name.lower() if self.name else repr(self)